import logging
from openhab_pythonrule_engine.item_registry import ItemRegistry
from openhab_pythonrule_engine.rule_engine import CronTrigger, ItemChangedTrigger, SystemEventTrigger, RuleEngine


def when(target: str):
    """
    Examples:
        .. code-block::
            @when("Time cron 55 55 5 * * ?")
            @when("Item gMotion_Sensors changed")
            @when("Rule loaded")
    Args:
        target (string): the `rules DSL-like formatted trigger expression <https://www.openhab.org/docs/configuration/rules-dsl.html#rule-triggers>`_
            to parse
    """

    if target.lower().startswith("item"):
        itemname_operation_pair = target[len("item"):].strip()
        itemname = itemname_operation_pair[:itemname_operation_pair.index(" ")].strip()

        if ItemRegistry.instance().has_item(itemname):
            operation = itemname_operation_pair[itemname_operation_pair.index(" "):].strip()

            def decorated_method(function):
                logging.info(" * " + function.__name__ + ": register trigger '" + target + "'")
                trigger = ItemChangedTrigger(itemname, operation, target, function)
                if trigger.is_valid():
                    RuleEngine.instance().add_item_changed_trigger(trigger)
                else:
                    logging.warning("Unsupported function spec " + str(function) + " Ignoring it")
                return function

            return decorated_method
        else:
            logging.warning("item " + itemname + " does not exist (trigger " + target + ")")


    elif target.lower().startswith("time cron"):
        cron = target[len("time cron"):].strip()

        def decorated_method(function):
            logging.info(" * " + function.__name__ + ": register cron    '" + target + "'")
            trigger = CronTrigger(cron, target, function)
            if trigger.is_valid():
                RuleEngine.instance().add_cron_trigger(trigger)
            else:
                logging.warning("Unsupported function spec " + str(function) + " Ignoring it")
            return function

        return decorated_method

    elif target.lower().strip() == "rule loaded":

        def decorated_method(function):
            logging.info(" * " + function.__name__ + ": register system  '" + target + "'")
            trigger = SystemEventTrigger(target, function)
            if trigger.is_valid():
                RuleEngine.instance().add_system_event_trigger(trigger)
            else:
                logging.warning("Unsupported function spec " + str(function) + " Ignoring it")
            return function

        return decorated_method

    else:
        logging.warning("unsupported expression " + target + " ignoring it")
        def decorated_method(function):
            return function
        return decorated_method



import logging
import requests
from datetime import datetime, time
from dataclasses import dataclass
from requests.auth import HTTPBasicAuth
from typing import Optional, List, Dict, Set, Any
from openhab_pythonrule_engine.cache import Cache


@dataclass
class Item:
    item_name: str
    read_only: bool
    group_names: List[str]
    value: Any

    def serialize(self, value) -> Optional[str]:
        pass

    def get_state(self):
        pass

    def get_state_as_text(self) -> str:
        pass

    def get_state_as_boolean(self) -> bool:
        pass

    def get_state_as_numeric(self) -> float:
        pass

    def get_state_as_datetime(self) -> datetime:
        text = self.get_state_as_text()
        try:
            # time_string may be in format 21:30
            return datetime.strptime(text, '%H:%M')
        except:
            try:
                # time_string may be in format "2017-07-14T21:30:00.000+0200"
                dt, timezone = text.split(".")
                return datetime.strptime(dt, '%Y-%m-%dT%H:%M:%S')
            except:
                try:
                    # time_string seems to be in format "2017-07-14T21:30:00"
                    return datetime.strptime(text, '%Y-%m-%dT%H:%M:%S')
                except:
                    return None


@dataclass
class TextItem(Item):
    value: str

    def get_state(self):
        return self.get_state_as_text()

    def get_state_as_text(self) -> str:
        return self.value

    def get_state_as_boolean(self) -> bool:
        if self.value is None:
            return False
        else:
            return True if (self.get_state_as_text() .lower() in ["true", "on"]) else False

    def get_state_as_numeric(self) -> float:
        if self.value is None:
            return -1
        else:
            return float(self.value)

    def serialize(self, value) -> Optional[str]:
        if self.value is None:
            return None
        elif type(value) == bool:
            return "ON" if value else "OFF"
        elif type(value) == datetime:
            return value.strftime('%Y-%m-%dT%H:%M:%S')
        else:
            return str(value)


@dataclass
class NumericItem(Item):
    item_name: str
    read_only: bool
    value: float

    def get_state(self):
        return self.get_state_as_numeric()

    def get_state_as_numeric(self) -> float:
        return self.value

    def get_state_as_boolean(self) -> bool:
        if self.value is None:
            return False
        else:
            return self.value != 0

    def get_state_as_text(self) -> str:
        if self.value is None:
            return ""
        return str(self.value)

    def serialize(self, value) -> Optional[str]:
        if self.value is None:
            return None
        else:
            if type(value) == bool:
                return "1" if value else "0"
            else:
                return str(value)


@dataclass
class BooleanItem(Item):
    item_name: str
    read_only: bool
    value: bool

    def get_state(self):
        return self.get_state_as_boolean()

    def get_state_as_boolean(self) -> bool:
        if self.value is None:
            return False
        else:
            return self.value

    def get_state_as_numeric(self) -> float:
        if self.value is None:
            return 0
        else:
            return 1 if self.value else 0

    def get_state_as_text(self) -> str:
        if self.value is None:
            return str(False)
        else:
            return str(self.value)

    def serialize(self, value) -> Optional[str]:
        if type(value) == bool:
            return "ON" if value else "OFF"
        elif type(value) == float:
            return "ON" if value == 1.0 else "OFF"
        elif type(value) == int:
            return "ON" if value == 1 else "OFF"
        else:
            return "ON" if (str(value).lower() in ["true", "on"]) else "OFF"


def to_item(data) -> Optional[Item]:
    try:
        if 'stateDescription' in data.keys():
            read_only = data['stateDescription']['readOnly']
        else:
            read_only = False
        if 'groupNames' in data.keys():
            group_names = data['groupNames']
        else:
            group_names = []
        state = data['state']
        if data['type'] == 'Number':
            item = NumericItem(data['name'], read_only, group_names, None if (state == 'NULL' or state == 'UNDEF') else float(state))
        elif data['type'] == 'Switch':
            item = BooleanItem(data['name'], read_only, group_names, None if state == 'NULL' else state == 'ON')
        else:
            item = TextItem(data['name'], read_only, group_names, None if (state == 'NULL' or state == 'UNDEF') else state)
        return item
    except Exception as e:
        logging.warning("error occurred mapping " + str(data) + " to item", e)
        return None




class ItemRegistry:
    __instance = None

    @staticmethod
    def new_singleton(openhab_uri: str, user: str, pwd: str):
        item_registry = ItemRegistry(openhab_uri, user, pwd)
        ItemRegistry.__instance = item_registry
        return item_registry

    @staticmethod
    def instance():
        return ItemRegistry.__instance

    def __init__(self, openhab_uri: str, user: str, pwd: str):
        self.cache = Cache()
        self.credentials = HTTPBasicAuth(user, pwd)
        if openhab_uri.endswith("/"):
            self.openhab_uri = openhab_uri
        else:
            self.openhab_uri = openhab_uri + "/"

    def on_event(self, event):
        if event.get("type", "") == "ThingUpdatedEvent":
            logging.info("config change. reset cache")
            self.cache.clear()

    def get_items(self, use_cache: bool = False) -> Dict[str, Item]:
        items = self.cache.read_entry("items", 24 * 60 * 60)
        if items is not None:
            return items
        else:
            uri = self.openhab_uri+ "rest/items"
            try:
                response = requests.get(uri, headers={"Accept": "application/json"}, auth = self.credentials)
                if response.status_code == 200:
                    items = {}
                    for entry in response.json():
                        item = to_item(entry)
                        if item is not None:
                            items[item.item_name] = item
                    self.cache.add_enry("items", items)
                    return items
                elif response.status_code == 404:
                    raise Exception("item " +   uri + " not exists " + response.text)
                else:
                    raise Exception("could not read item state " +   uri +  " got error " + response.text)
            except Exception as e:
                logging.warning("error occurred by calling " + uri, e)

    def get_item(self, item_name: str) -> Optional[Item]:
        uri = self.openhab_uri+ "rest/items/" + item_name
        try:
            response = requests.get(uri, headers={"Accept": "application/json"}, auth = self.credentials)
            if response.status_code == 200:
                data = response.json()
                return to_item(data)
            elif response.status_code == 404:
                raise Exception("item " +   uri + " not exists " + response.text)
            else:
                raise Exception("could not read item state " +   uri +  " got error " + response.text)
        except Exception as e:
            logging.warning("error occurred by calling " + uri, e)

    def has_item(self, item_name: str) -> bool:
        return self.get_item(item_name) != None

    def get_group_membernames(self, group_name) -> List[str]:
        return [item.item_name for item in self.get_items().values() if group_name in item.group_names]

    def set_item_state(self, item_name: str, value: str):
        uri = self.openhab_uri+ "rest/items/" + item_name
        try:
            response = requests.post(uri, data=value, headers={"Content-type": "text/plain"}, auth = self.credentials)
            if response.status_code == 200:
                return
            elif response.status_code == 404:
                raise Exception("item " +   uri + " not exists " + response.text)
            else:
                raise Exception("could not update item state " +   uri +  " got error " + response.text)
        except Exception as e:
            logging.warning("error occurred by performing put on " + uri, e)

    def get_item_metadata(self, item_name: str) -> Optional[Item]:
        items_meta_data = self.get_items(use_cache=True)
        for name in items_meta_data.keys():
            if item_name == name:
                return items_meta_data[item_name]
        return None

    def get_state(self, item_name: str, dflt):
        state = self.get_item(item_name)
        if state is None or state.value is None:
            return dflt
        else:
            return state.get_state()

    def get_state_as_numeric(self, item_name: str, dflt: float=-1) -> float:
        state = self.get_item(item_name)
        if state is None or state.value is None:
            return dflt
        else:
            return state.get_state_as_numeric()

    def get_state_as_boolean(self, item_name: str, dflt: bool=False) -> bool:
        state = self.get_item(item_name)
        if state is None or state.value is None:
            return dflt
        else:
            return state.get_state_as_boolean()

    def get_state_as_text(self, item_name: str, dflt: str="") -> str:
        state = self.get_item(item_name)
        if state is None or state.value is None:
            return dflt
        else:
            return state.get_state_as_text()

    def get_state_as_datetime(self, item_name: str, datetime_string: str="1970-01-01") -> datetime:
        state = self.get_item(item_name)
        if state is None or state.value is None:
            return self.__to_datetime_object(datetime_string)
        else:
            return state.get_state_as_datetime()

    def __to_datetime_object(self, datetime_string: str):
        try:
            # time_string may be in format 21:30
            return datetime.strptime(datetime_string, '%H:%M')
        except:
            try:
                # time_string may be in format "2017-07-14T21:30:00.000+0200"
                dt, timezone = datetime_string.split(".")
                return datetime.strptime(dt, '%Y-%m-%dT%H:%M:%S')
            except:
                # time_string seems to be in format "2017-07-14T21:30:00"
                return datetime.strptime(datetime_string, '%Y-%m-%dT%H:%M:%S')

    def set_state(self, item_name: str, value):
        if value is None:
            logging.warning("try to set " + item_name + " = None. ignoring it")
            return
        else:
            item_metadata = self.get_item_metadata(item_name)
            if item_metadata is None:
                raise Exception("item " + item_name + " not exists")
            else:
                new_state = item_metadata.serialize(value)
                if self.get_state(item_name, None) != new_state:
                    self.set_item_state(item_name, item_metadata.serialize(value))
                    logging.debug("set " + item_name + " = " + new_state)


'''
logging.basicConfig(format='%(asctime)s %(name)-20s: %(levelname)-8s %(message)s', level='INFO', datefmt='%Y-%m-%d %H:%M:%S')

item_registry = ItemRegistry("http://localhost:8080/", "grro", "Stabilo33!")

item_registry.set_state("Volumio_playing", False)
print(" ")


item_registry.set_state("InternetConnectivityMonitor_Updatedtime", datetime.now())


print(item_registry.get_state_as_datetime("InternetConnectivityMonitor_Updatedtime"))

item_registry.set_state("InternetConnectivityMonitor_Updatedtime", datetime.now())



print(item_registry.get_state_as_boolean("WindSensor_Windspeed"))

item_registry.set_state("Volumio_playing", "OFF")
print(item_registry.get_state_as_boolean("Volumio_playing"))

item_registry.set_state("Volumio_playing", True)
print(item_registry.get_state_as_boolean("Volumio_playing"))

item_registry.set_state("Volumio_playing", 0)
print(item_registry.get_state_as_boolean("Volumio_playing"))

item_registry.set_state("Volumio_playing", "ON")
print(item_registry.get_state_as_boolean("Volumio_playing"))

item_registry.set_state("Volumio_playing", False)
print(item_registry.get_state_as_boolean("Volumio_playing"))

item_registry.set_state("Volumio_playing", 1)
print(item_registry.get_state_as_boolean("Volumio_playing"))

item_registry.set_state("Volumio_playing", 0)



logging.basicConfig(format='%(asctime)s %(name)-20s: %(levelname)-8s %(message)s', level='INFO', datefmt='%Y-%m-%d %H:%M:%S')

item_registry = ItemRegistry("http://localhost:8080/", "grro", "Stabilo33!")

print(item_registry.get_datetime("InternetConnectivityMonitor_Updatedtime"))
print(item_registry.get_boolean("InternetConnectivityMonitor_Updatedtime"))
#print(item_registry.get_numeric("InternetConnectivityMonitor_Updatedtime"))
print(item_registry.get_text("InternetConnectivityMonitor_Updatedtime"))


print(item_registry.get_boolean("Volumio_title"))
#print(item_registry.get_numeric("Volumio_title"))
print(item_registry.get_text("Volumio_title"))
#print(item_registry.get_datetime("Volumio_title"))
print(" ")

print(item_registry.get_boolean("Volumio_playing"))
print(item_registry.get_numeric("Volumio_playing"))
print(item_registry.get_text("Volumio_playing"))
print(" ")


print(item_registry.get_boolean("WindSensor_Windspeed"))
print(item_registry.get_numeric("WindSensor_Windspeed"))
print(item_registry.get_text("WindSensor_Windspeed"))
print(" ")
'''


import logging
from abc import ABC
from datetime import datetime
from dataclasses import dataclass
from typing import Optional
from openhab_pythonrule_engine.invoke import Invoker
from openhab_pythonrule_engine.item_registry import ItemRegistry


@dataclass
class Execution:
    datetime: datetime
    error: Optional[Exception] = None

    def __str__(self):
        text = self.datetime.strftime("%Y-%m-%d-T%H:%M:%S")
        if self.error is not None:
            text += "  (Error: " + str(self.error) + ")"
        return text

    def __repr__(self):
        return self.__str__()


class Trigger(ABC):

    def __init__(self, expression: str, func):
        self.expression = expression
        self.func = func
        self.invoker = Invoker.create(func)
        self.last_executions = []
        self.last_errors = {}

    def is_valid(self):
        return self.invoker is not None

    def invoke(self, item_registry: ItemRegistry):
        execution = None
        if len(self.last_executions) > 20:
            self.last_executions.pop(0)
        try:
            self.invoker.invoke(item_registry)
            self.last_executions.append(Execution(datetime.now(), None))
        except Exception as e:
            self.last_executions.append(Execution(datetime.now(), e))
            logging.warning("Error occurred by invoking " + self.name, e)



    @property
    def module(self):
        return self.func.__module__

    @property
    def name(self):
        return self.func.__name__

    def __str__(self):
        return self.expression


class CronTrigger(Trigger):

    def __init__(self, cron: str, expression: str, func):
        self.cron = cron
        super().__init__(expression, func)


class SystemEventTrigger(Trigger):

    def __init__(self, expression: str, func):
        super().__init__(expression, func)


class ItemChangedTrigger(Trigger):

    def __init__(self, item_name: str, operation: str, expression: str, func):
        self.item_name = item_name
        self.operation = operation
        super().__init__(expression, func)

    def on_event(self, event):
        topic = event.get("topic", "")
        if topic.startswith('openhab') or topic.startswith('smarthome'):
            try:
                parts = topic.split("/")
                #print(parts)
                if parts[1] == 'items':
                    item_name = parts[2]
                    if item_name == self.item_name:
                        operation = parts[3]
                        if operation == 'statechanged':
                            logging.debug("executing rule " + self.invoker.name + " (trigger 'Item " + item_name + " changed')")
                            self.invoke(ItemRegistry.instance())
            except Exception as e:
                logging.warning("Error occurred by handling event " + str(event), e)



class TriggerRegistry:

    def __init__(self):
        self.triggers_by_module = {}

    def register(self, trigger: Trigger):
        triggers = self.triggers_by_module.get(trigger.module, set())
        triggers.add(trigger)
        self.triggers_by_module[trigger.module] = triggers

    def deregister(self, module):
        if module in list(self.triggers_by_module.keys()):
            logging.info("removing all " + str(len(self.triggers_by_module[module])) + " triggers of '" + module + "'")
            del self.triggers_by_module[module]

    def get_triggers_by_type(self, type):
        filterred_triggers = []
        for triggers in self.triggers_by_module.values():
            for trigger in triggers:
                if isinstance(trigger, type):
                    filterred_triggers.append(trigger)
        return filterred_triggers




from webthing import (SingleThing, Value, Property, Thing, MultipleThings, WebThingServer)
import tornado.ioloop
import logging
from openhab_pythonrule_engine.rule_engine import RuleEngine, Rule


class RuleThing(Thing):

    def __init__(self, description: str, rule: Rule):
        Thing.__init__(
            self,
            'urn:dev:ops:pythonrule-1',
            'python_rule',
            [],
            description
        )

        self.name = Value(rule.module + "." + rule.name)
        self.add_property(
            Property(self,
                     'name',
                     self.name,
                     metadata={
                         'title': 'name',
                         'type': 'string',
                         'description': 'the rule name',
                         'readOnly': True
                     }))

        self.ioloop = tornado.ioloop.IOLoop.current()
        



def run_server(port: int, description: str, rule_engine: RuleEngine):
    rule_webthings = [RuleThing(description, rule) for rule in rule_engine.rules]
    server = WebThingServer(MultipleThings(rule_webthings, 'rule'), port=port, disable_host_validation=True)

    try:
        # start webthing server
        logging.info('starting the server listing on ' + str(port))
        server.start()
    except KeyboardInterrupt:
        logging.info('stopping the server')
        server.stop()
        logging.info('done')




import logging
import os
import sys
import importlib
import pycron
from time import sleep
from threading import Thread
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from typing import List
from openhab_pythonrule_engine.item_registry import ItemRegistry
from openhab_pythonrule_engine.trigger import TriggerRegistry, CronTrigger, ItemChangedTrigger, SystemEventTrigger, Trigger
from openhab_pythonrule_engine.eventbus_consumer import EventConsumer


class FileSystemListener(FileSystemEventHandler):

    def __init__(self, rule_engine):
        self.rule_engine = rule_engine

    def on_moved(self, event):
        self.rule_engine.unload_module(self.filename(event.src_path))
        self.rule_engine.load_module(self.filename(event.dest_path))

    def on_deleted(self, event):
        self.rule_engine.unload_module(self.filename(event.src_path))

    def on_created(self, event):
        self.rule_engine.load_module(self.filename(event.src_path))

    def on_modified(self, event):
        self.rule_engine.load_module(self.filename(event.src_path))

    def filename(self, path):
        path = path.replace("\\", "/")
        return path[path.rindex("/")+1:]


class CronScheduler:

    def __init__(self):
        self.cron_trigger_by_module = {}
        self.thread = Thread(target=self.__process, daemon=True)

    def add_job(self, cron_trigger: CronTrigger):
        cron_triggers = self.cron_trigger_by_module.get(cron_trigger.module, set())
        cron_triggers.add(cron_trigger)
        self.cron_trigger_by_module[cron_trigger.module] = cron_triggers

    def remove_jobs(self, module: str):
        if module in self.cron_trigger_by_module.keys():
            logging.info("removing all " + str(len(self.cron_trigger_by_module[module])) + " crons of '" + module + "'")
            del self.cron_trigger_by_module[module]

    def __process(self):
        while True:
            for cron_triggers in list(self.cron_trigger_by_module.values()):
                for cron_trigger in list(cron_triggers):
                    if pycron.is_now(cron_trigger.cron):
                        try:
                            logging.debug("executing rule " + cron_trigger.name + " (cron 'Time cron " + cron_trigger.cron + "')")
                            cron_trigger.invoke(ItemRegistry.instance())
                        except Exception as e:
                            logging.warning("Error occurred by executing rule " + cron_trigger.name, e)
            sleep(60)



class Rule:

    def __init__(self, func):
        self.func = func
        self.triggers = []

    @property
    def module(self):
        return self.func.__module__

    @property
    def name(self):
        return self.func.__name__

    def __str__(self):
        text = self.module + ".py#" + self.name
        for trigger in self.triggers:
            text += "\r\n  * " + trigger.expression
            for execution in trigger.last_executions:
                text += "\r\n      * " + str(execution)

        return  text

    def __repr__(self):
        return self.__str__()

class RuleEngine:

    __instance = None

    @staticmethod
    def instance():
        return RuleEngine.__instance

    @staticmethod
    def start_singleton(openhab_uri:str, python_rule_directory: str = "/etc/openhab/automation/rules", user: str = None, pwd: str = None):
        rule_engine = RuleEngine(openhab_uri, python_rule_directory, user, pwd)
        RuleEngine.__instance = rule_engine
        rule_engine.start()

    def __init__(self, openhab_uri:str, python_rule_directory: str, user: str, pwd: str):
        self.__python_rule_directory = python_rule_directory
        logging.info("connecting " + openhab_uri)
        ItemRegistry.new_singleton(openhab_uri, user, pwd)
        self.__event_consumer = EventConsumer(openhab_uri, self)
        self.cron_scheduler = CronScheduler()
        self.trigger_registry = TriggerRegistry()

    def start(self):
        logging.info("rules directory is " + self.__python_rule_directory)
        if self.__python_rule_directory not in sys.path:
            sys.path.insert(0, self.__python_rule_directory)

        for file in os.scandir(self.__python_rule_directory):
            self.load_module(file.name)

        observer = Observer()
        observer.schedule(FileSystemListener(self), self.__python_rule_directory, recursive=False)
        observer.start()

        self.__event_consumer.start()

    def add_cron_trigger(self, trigger: CronTrigger):
        self.trigger_registry.register(trigger)
        self.cron_scheduler.add_job(trigger)

    def add_item_changed_trigger(self, trigger: ItemChangedTrigger):
        self.trigger_registry.register(trigger)

    def add_system_event_trigger(self, trigger: SystemEventTrigger):
        trigger.invoke(ItemRegistry.instance())
        self.trigger_registry.register(trigger)

    def load_module(self, filename: str):
        if filename.endswith(".py"):
            modulename = self.__filename_to_modulename(filename)
            # reload?
            if modulename in sys.modules:
                logging.info("reload '" + filename + "'")
                self.cron_scheduler.remove_jobs(modulename)
                self.trigger_registry.deregister(modulename)
                importlib.reload(sys.modules[modulename])
            else:
                logging.info("load '" + filename + "'")
                importlib.import_module(modulename)

    def unload_module(self, filename: str):
        modulename = self.__filename_to_modulename(filename)
        if modulename in sys.modules:
            logging.info("unload '" + filename + "'")
            self.cron_scheduler.remove_jobs(modulename)
            self.trigger_registry.deregister(modulename)
            del sys.modules[modulename]

    def __filename_to_modulename(self, filename):
        return filename[:-3]

    def on_event(self, event):
        ItemRegistry.instance().on_event(event)
        for item_changed_trigger in self.trigger_registry.get_triggers_by_type(ItemChangedTrigger):
            item_changed_trigger.on_event(event)

    @property
    def rules(self) -> List[Rule]:
        rules = []
        for module in self.trigger_registry.triggers_by_module.keys():
            triggers = self.trigger_registry.triggers_by_module[module]
            for func in [trigger.func for trigger in triggers]:
                rule = Rule(func)
                rules.append(rule)
                for trigger in triggers:
                    if trigger.name == rule.name:
                        rule.triggers.append(trigger)
        return rules